#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <sys/stat.h>

/* read arpRofs.cvm (from Arcana Heart for PS2) */

/* currently needs to be given the beginning of a directory, and will
   only extract that directory */

off_t next_dir=0xa4800;
char curdir[PATH_MAX+1];

unsigned long read32(unsigned char * buf) {
    return buf[0]|(buf[1]<<8)|(buf[2]<<16)|(buf[3]<<24);
}

struct fileinfo {
    char name[0x20];
    unsigned long size;
    unsigned long offset;
};

#define DUMP_BUFFER_SIZE 2048
char buf[DUMP_BUFFER_SIZE];
int dump_file(FILE * infile, off_t offset, size_t size, const char * filename) {
    FILE * outfile;
    off_t oldoff=ftell(infile);

    outfile = fopen(filename,"wb");

    if (!outfile) {
        fprintf(stderr,"error opening %s for output\n",filename);
        return 1;
    }

    if (fseek(infile,offset,SEEK_SET)) {
        fprintf(stderr,"seek error\n");
        fseek(infile,oldoff,SEEK_SET);
        return 1;
    }

    while (size>=DUMP_BUFFER_SIZE) {
        size-=DUMP_BUFFER_SIZE;
        if (fread(buf,1,DUMP_BUFFER_SIZE,infile)!=DUMP_BUFFER_SIZE ||
            fwrite(buf,1,DUMP_BUFFER_SIZE,outfile)!=DUMP_BUFFER_SIZE) {
            fprintf(stderr,"error dumping\n");
            fclose(outfile);
            return 1;
        }
    }
    if (size) {
        if (fread(buf,1,size,infile)!=size ||
                fwrite(buf,1,size,outfile)!=size) {
            fprintf(stderr,"error dumping\n");
            fclose(outfile);
            return 1;
        }
    }

    fseek(infile,oldoff,SEEK_SET);

    return 0;
}

int read_directory(FILE * infile, off_t offset, int dump) {
    int count;
    struct fileinfo fileinfo;
    off_t fileoff, temp;
    off_t oldoff=ftell(infile); /* return the pointer here when done */
    int i;

    if (fseek(infile,offset,SEEK_SET)) {
        fprintf(stderr,"Failed to read directory at 0x%08x, seek failed\n",(unsigned int)offset);
        fseek(infile,oldoff,SEEK_SET);
        return 1;
    }

    if (fread(&fileinfo,sizeof(struct fileinfo),1,infile) != 1) {
        fprintf(stderr,"Failed to read directory at 0x%08x\n",(unsigned int)offset);
        fseek(infile,oldoff,SEEK_SET);
        return 1;
    }
    for (i=0;i<0x20;i++) {
        if (fileinfo.name[i]) {
            fprintf(stderr,"Directory should start with an empty name field,\n\
                    something is probably wrong\n");
            fseek(infile,oldoff,SEEK_SET);
            return 1;
        }
    }
    count = read32((unsigned char *)(&fileinfo.size));
    fileoff = read32((unsigned char *)(&fileinfo.offset));

    printf("%d files, directory size 0x%08x\n",count,(unsigned int)fileoff);
    for (i=0;i<count;i++) {
        if (fread(&fileinfo,sizeof(struct fileinfo),1,infile) != 1) {
            fprintf(stderr,"Failed to read info for file %d in directory at %08x\n",i,(unsigned int)offset);
            fseek(infile,oldoff,SEEK_SET);
            return 1;
        }

        fileinfo.size=read32((unsigned char *)(&fileinfo.size));
        fileinfo.offset=read32((unsigned char *)(&fileinfo.offset));
        printf("%d\t%s\tsize:%08x\toffset:%08x (%08x)\n",i,fileinfo.name,(unsigned int)fileinfo.size,(unsigned int)(fileinfo.offset+fileoff),(unsigned int)fileinfo.offset);
        if (dump) {
            printf("dumping %s\n",fileinfo.name);
            if (dump_file(infile,offset+fileoff+fileinfo.offset,fileinfo.size,fileinfo.name))
                return 1;
        }
    }

    fseek(infile,oldoff,SEEK_SET);

    next_dir=((offset+fileoff+fileinfo.offset+fileinfo.size)+0x7ff)/0x800*0x800;
    printf("Probable next directory: %08x\n",next_dir);

    return 0;
}

int main(int argc, char ** argv) {
    FILE * infile = fopen("arpRofs.cvm","rb");
    int dump=0;
    /* TODO: check header */
    printf("arpRofs extractor 0.0\n");
    if (!infile) {
        fprintf(stderr,"I'm looking for a friend of mine, you might know him,\ngoes by \"arpRofs.cvm\"\n");
        return -1;
    }

    getcwd(curdir,sizeof(curdir)-1);

    do {
        char buf[100+1]={};
        off_t next_dir_save=next_dir;
        chdir(curdir);
        printf("where to? (%x) ",next_dir);
        fgets(buf,100,stdin);
        if (sscanf(buf,"%x",&next_dir)!=1) next_dir=next_dir_save;

        memset(buf,0,sizeof(buf));
        printf("dump? (%c) ",dump?'y':'n');
        fgets(buf,100,stdin);
        if (buf[0]=='y' || buf[0]=='Y') dump=1;
        if (buf[0]=='n' || buf[0]=='N') dump=0;

        if (dump) {
            /* make up a directory */
            sprintf(buf,"dir_%08x",next_dir);
#ifdef __MINGW_H
            mkdir(buf);
#else
            mkdir(buf,0744);
#endif
            chdir(buf);
        }


    } while (!read_directory(infile,next_dir,dump));

    chdir(curdir);

    return 0;
}
